﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TestFunctions.MultiObjective;
namespace HeuristicLab.Algorithms.MOCMAEvolutionStrategy {
  [Item("HypervolumeIndicator", "Selection of Offspring based on contributing Hypervolume")]
  [StorableClass]
  class HypervolumeIndicator : Item, IIndicator {

    public int LeastContributer<R>(R[] front, Func<R, double[]> extractor, MultiObjectiveTestFunctionProblem problem) {
      DoubleMatrix bounds = problem.Bounds;
      int mindex = 1;
      double min = Double.MaxValue;
      List<double[]> frontArr = front.Select(x => extractor.Invoke(x)).ToList<double[]>();
      double[] refPoint = buildReference(frontArr, problem.Maximization);
      refPoint = buildReference(refPoint, problem.TestFunction.ReferencePoint(problem.Objectives), problem.Maximization);
      double hv = Hypervolume.Calculate(frontArr, refPoint, problem.Maximization);
      for (int i = 0; i < frontArr.Count; i++) {
        var d = Contribution(frontArr, i, problem.Maximization, refPoint, hv);
        if (min > d) {
          mindex = i;
          min = d;
        }
      }

      return mindex;
    }

    private double Contribution(List<double[]> front, int idx, bool[] maximization, double[] refPoint, double hv) {
      var point = front[idx];
      front.RemoveAt(idx);
      var contribution = hv - Hypervolume.Calculate(front, refPoint, maximization);
      front.Insert(idx, point);
      return contribution;
    }

    private double[] buildReference(IEnumerable<double[]> front, bool[] maximization) {
      double[] refPoint = new double[maximization.Length];
      foreach (double[] point in front) {
        for (int i = 0; i < maximization.Length; i++) {
          refPoint[i] = maximization[i] ? Math.Min(refPoint[i], point[i]) : Math.Max(refPoint[i], point[i]);
        }
      }
      return refPoint;
    }
    private double[] buildReference(double[] point1, double[] point2, bool[] maximization) {
      double[] refPoint = new double[maximization.Length];

      for (int i = 0; i < maximization.Length; i++) {
        refPoint[i] = maximization[i] ? Math.Min(point2[i], point1[i]) : Math.Max(point2[i], point1[i]);
      }

      return refPoint;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new HypervolumeIndicator(this, cloner);
    }

    [StorableConstructor]
    protected HypervolumeIndicator(bool deserializing) : base(deserializing) { }
    protected HypervolumeIndicator(HypervolumeIndicator original, Cloner cloner)
      : base(original, cloner) {
    }
    public HypervolumeIndicator() { }
  }
}
